u"""
Yet Another Photon Classifier (YAPC) for Python
===============================================

yapc contains Python tools for classifying geolocated photon data
from the NASA Ice, Cloud and land Elevation Satellite-2 (ICESat-2)
using a k-Nearest Neighbors (kNN) algorithm

The package works using Python packages (numpy, scikit-learn)

Documentation is available at https:/yapc.readthedocs.io
"""
import yapc._dist_metrics
import yapc.utilities
import yapc.version
from yapc.classify_photons import classify_photons
# set version
__version__ = yapc.version.version



#!/usr/bin/env python
u"""
classify_photons.py
Written by Aimee Gibbons and Tyler Sutterley (06/2022)
Yet Another Photon Classifier for ATL03 Geolocated Photon Data

PYTHON DEPENDENCIES:
    numpy: Scientific Computing Tools For Python
        https://numpy.org
        https://numpy.org/doc/stable/user/numpy-for-matlab-users.html
    scipy: Scientific Tools for Python
        https://docs.scipy.org/doc/
    scikit-learn: Machine Learning in Python
        http://scikit-learn.org/stable/index.html
        https://github.com/scikit-learn/scikit-learn

UPDATE HISTORY:
    Updated 06/2022: added option for setting the minimum KNN value
    Updated 04/2022: can weight using only height differences
    Updated 10/2021: half the perimeter for weighting the distances
        scale weights by the selected number of neighbors (K)
        check that the telemetry band height is positive
    Updated 09/2021: add option for setting aspect ratio of window
        add option to return selection window dimensions
    Updated 08/2021: update algorithm to match current GSFC version
    Updated 05/2021: use int64 to fix numpy deprecation warning
    Written 05/2021
"""
import numpy as np
import sklearn.neighbors
import yapc._dist_metrics as _dist_metrics

# PURPOSE: create distance metric for windowed classifier
def windowed_manhattan(u, v, window=[], w=[]):
    """
    Create a windowed Manhattan distance metric

    Parameters
    ----------
    u: float
        Input array
    v: float
        Input array for distance
    window: float or list, default []
        distance window for reducing neighbors
    w: float or list, default []
        weights for each value
    """
    # verify dimensions
    u = np.atleast_1d(u)
    v = np.atleast_1d(v)
    # calculate Manhattan (rectilinear) distances
    l1_diff = np.abs(u - v)
    # broadcast window to dimensions if using a square window
    window = np.broadcast_to(np.atleast_1d(window),l1_diff.shape)
    w = np.broadcast_to(np.atleast_1d(w),l1_diff.shape)
    for d,wnd in enumerate(window):
        if (l1_diff[d] >= wnd):
            l1_diff[d] = np.inf
        # scale differences by weights
        with np.errstate(invalid='ignore'):
            l1_diff[d] *= w[d]
    return l1_diff.sum()

# PURPOSE: calculate distances between points as matrices
def distance_matrix(u, v, p=1, window=[], w=[]):
    """
    Calculate distances between two collections of points

    Parameters
    ----------
    u: float
        First collection of coordinates
    v: float
        Second collection of coordinates
    p: int, default 1
        Power for calculating distance

            - ``1``: Manhattan distances
            - ``2``: Euclidean distances
    window: float or list, default []
        Distance window for reducing neighbors
    w: float or list, default []
        weights for each value
    """
    M,s = np.shape(u)
    N,s = np.shape(v)
    # allocate for output distance matrix
    D = np.zeros((M,N))
    # broadcast window to dimensions if using a square window
    window = np.broadcast_to(np.atleast_1d(window),(s,))
    w = np.broadcast_to(np.atleast_1d(w),(s,))
    for d in range(s):
        ii, = np.dot(d,np.ones((1,N))).astype(np.int64)
        jj, = np.dot(d,np.ones((1,M))).astype(np.int64)
        dx = np.abs(u[:,ii] - v[:,jj].T)
        # window differences for dimension
        dx[dx >= window[d]] = np.inf
        # scale differences by weights
        # add differences to total distance matrix
        with np.errstate(invalid='ignore'):
            D += np.power(w[d]*dx, p)
    # convert distances to output units
    return np.power(D, 1.0/p)

# PURPOSE: use the GSFC YAPC k-nearest neighbors algorithm to determine
# weights for each photon event
def classify_photons(x, h, h_win_width, indices, **kwargs):
    """
    Use the NASA GSFC YAPC k-nearest neighbors algorithm to determine
    weights for each photon event

    Parameters
    ----------
    x: float
        along-track x coordinates for photon events
    h: float
        photon event heights
    h_win_width: float
        height of (possibly 2) telemetry bands
    indices: int
        indices of photon events to classify
    K: int, default 0
        number of values for KNN algorithm
    min_knn: int, default 5
        minimum number of values for KNN algorithm
    min_ph: int, default 3
        minimum number of photons to be valid
    min_xspread: float, default 1.0
        minimum along-track spread of photon events
    min_hspread: float, default 0.01
        minimum window of heights for photon events
    win_x: float, default 15.0
        along-track length of window
    win_h: float, default 6.0
        height of window
    aspect: float, default 0.0
        aspect ratio of x and h window
    method: str, default 'linear'
        algorithm for computing photon event weights

            - ``'ball_tree'``: use scikit.learn.BallTree with custom distance metric
            - ``'linear'``: use a brute-force approach with linear algebra
            - ``'brute'``: use a brute-force approach
    metric: str, default 'height'
        metric for computing distances

            - ``'height'``: height differences
            - ``'manhattan'``: manhattan distances
    return_window: bool, default False
        return the width and height of the selection window
    return_K: bool, default False
        return the dynamically selected number of values
    """
    # set default keyword arguments
    kwargs.setdefault('K', 0)
    kwargs.setdefault('min_knn', 5)
    kwargs.setdefault('min_ph', 3)
    kwargs.setdefault('min_xspread', 1.0)
    kwargs.setdefault('min_hspread', 0.01)
    kwargs.setdefault('win_x', 15.0)
    kwargs.setdefault('win_h', 6.0)
    kwargs.setdefault('aspect', 0.0)
    kwargs.setdefault('method', 'linear')
    kwargs.setdefault('metric', 'height')
    kwargs.setdefault('return_window', False)
    kwargs.setdefault('return_K', False)
    # number of photon events
    n_pe = len(h[indices])
    # output photon weights
    pe_weights = np.zeros((n_pe))
    # number of values for KNN algorithm
    if (kwargs['K'] == 0):
        K = np.max([kwargs['min_knn'], np.sqrt(n_pe)/2]).astype(int)
    else:
        K = np.copy(kwargs['K'])
    # check that number of photons is greater than criteria
    # number of points but be greater than or equal to k
    min_ph_check = (n_pe >= kwargs['min_ph']) & (n_pe >= (K+1))
    if np.logical_not(min_ph_check) and kwargs['return_window'] and kwargs['return_K']:
        # return empty weights, window sizes and selection window
        return (pe_weights, 0.0, 0.0, K)
    if np.logical_not(min_ph_check) and kwargs['return_window']:
        # return empty weights and window sizes
        return (pe_weights, 0.0, 0.0)
    if np.logical_not(min_ph_check) and kwargs['return_K']:
        # return empty weights and selection window
        return (pe_weights, K)
    elif np.logical_not(min_ph_check):
        # return empty weights
        return pe_weights
    # along-track spread of photon events
    xspread = np.max(x[indices]) - np.min(x[indices])
    # height spread of photon events
    hspread = np.max(h[indices]) - np.min(h[indices])
    # check that spread widths are greater than criteria
    spread_check = (xspread >= kwargs['min_xspread']) & \
        (hspread >= kwargs['min_hspread']) & \
        (h_win_width >= 0.0)
    if np.logical_not(spread_check) and kwargs['return_window'] and kwargs['return_K']:
        # return empty weights, window sizes and selection window
        return (pe_weights, 0.0, 0.0, K)
    if np.logical_not(spread_check) and kwargs['return_window']:
        # return empty weights and window sizes
        return (pe_weights, 0.0, 0.0)
    if np.logical_not(spread_check) and kwargs['return_K']:
        # return empty weights and selection window
        return (pe_weights, K)
    elif np.logical_not(spread_check):
        # return empty weights
        return pe_weights
    # use pre-defined window size or adaptive window size
    if (kwargs['aspect'] > 0.0):
        # photon density
        density = n_pe/(xspread*h_win_width)
        # minimum area to contain minimum number of photon events
        area_min = kwargs['min_ph']/density
        # minimum length of a square containing minimum number of photons
        length_min = np.sqrt(area_min)
        # calculate horizontal and vertical window sizes with aspect ratio
        # x window length will be aspect times the h window length
        win_x = np.power(kwargs['aspect'], 0.5)*length_min
        win_h = np.power(kwargs['aspect'], -0.5)*length_min
    else:
        # pre-defined window size
        win_x = np.copy(kwargs['win_x'])
        win_h = np.copy(kwargs['win_h'])
    # reduce to a buffered window
    xmin = np.min(x[indices]) - win_x/2.0
    xmax = np.max(x[indices]) + win_x/2.0
    hmin = np.min(h[indices]) - win_h/2.0
    hmax = np.max(h[indices]) + win_h/2.0
    iwin, = np.nonzero((x >= xmin) & (x <= xmax) & (h >= hmin) & (h <= hmax))
    # normalization for weights
    if (kwargs['metric'] == 'height'):
        dist_norm = K*win_h/2.0
        dist_weights = np.array([0.0, 1.0])
    else:
        dist_norm = K*(win_x/2.0 + win_h/2.0)
        dist_weights = np.array([1.0, 1.0])
    # method of calculating photon event weights
    if (kwargs['method'] == 'ball_tree'):
        # use BallTree with custom metric to calculate photon event weights
        # window for nearest neighbors
        window = np.array([win_x/2.0,win_h/2.0])
        # create ball tree with photon events
        # using a cythonized callable distance metric
        tree = sklearn.neighbors.BallTree(np.c_[x[iwin],h[iwin]],
            metric=_dist_metrics.windowed_manhattan, window=window,
            w=dist_weights)
        # K nearest neighbors with windowed manhattan metric
        # use K+1 to remove identity distances (d=0)
        dist,_ = tree.query(np.c_[x[indices],h[indices]], k=K+1,
            return_distance=True)
        # calculate photon event weights and normalize by window size
        valid = np.all(np.isfinite(dist),axis=1)
        inv_dist = np.sum(win_x/2.0 + win_h/2.0 - dist[:,1:],axis=1)
        pe_weights[valid] = inv_dist[valid]/dist_norm
    elif (kwargs['method'] == 'linear'):
        # use brute force with linear algebra to calculate photon event weights
        # window for nearest neighbors
        window = np.array([win_x/2.0,win_h/2.0])
        # calculate distance matrix between points
        dist = distance_matrix(np.c_[x[indices],h[indices]],
            np.c_[x[iwin],h[iwin]], p=1, window=window, w=dist_weights)
        # sort distances and get K nearest neighbors
        # use K+1 to remove identity distances (d=0)
        dist_sort = np.sort(dist, axis=1)[:,1:K+1]
        # calculate inverse distance of photon events in window
        inv_dist = win_x/2.0 + win_h/2.0 - dist_sort
        # calculate photon event weights and normalize by window size
        valid = np.all(np.isfinite(dist_sort),axis=1)
        pe_weights[valid] = np.sum(inv_dist[valid,:], axis=1)/dist_norm
    elif (kwargs['method'] == 'brute'):
        # use brute force approach to calculate photon event weights
        for j,i in enumerate(indices):
            # all photon events in buffer excluding source photon
            ii = sorted(set(iwin) - set([i]))
            # distance of photon events to source photon
            dx = np.abs(x[ii] - x[i])
            dh = np.abs(h[ii] - h[i])
            # indices of photons within window
            n, = np.nonzero((dx < (win_x/2.0)) & (dh < (win_h/2.0)))
            # skip iteration if there are less than K within window
            if (len(n) < K):
                continue
            # calculate inverse distance of photon events in window
            # sort distances and get K nearest neighbors
            if (kwargs['metric'] == 'height'):
                inv_dist = win_h/2.0 - dh[n]
                k_sort = np.argsort(dh[n])[:K]
            else:
                inv_dist = win_x/2.0 - dx[n] + win_h/2.0 - dh[n]
                k_sort = np.argsort(dx[n] + dh[n])[:K]
            # sum of the K largest weights (normalized by the window size)
            pe_weights[j] = np.sum(inv_dist[k_sort])/dist_norm
    # check if returning both the weights and auxiliary data
    if kwargs['return_window'] and kwargs['return_K']:
        # return the weights, window size and dynamically selected K
        return (pe_weights, win_x, win_h, K)
    elif kwargs['return_window']:
        # return the weights and window size
        return (pe_weights, win_x, win_h)
    elif kwargs['return_K']:
        # return the weights and dynamically selected K
        return (pe_weights, K)
    else:
        # return the weights
        return pe_weights



#!python
#cython: boundscheck=False
#cython: wraparound=False
#cython: cdivision=True

# Inline distance functions

from libc.math cimport fabs, INFINITY

# check if distance is within window
cdef inline double in_window(double d, double w):
    return d if d < w else INFINITY

# calculate windowed manhattan (rectilinear) distances
cdef inline dist(double[:] u, double[:] v, double[:] window, double[:] w):
    cdef int n = u.shape[0]
    cdef double d = 0.0
    for j in range(n):
        d += w[j]*in_window(fabs(u[j] - v[j]), window[j])
    return d



#!/usr/bin/env python
u"""
utilities.py
Written by Tyler Sutterley (10/2021)
Management utilities for Yet Another Photon Classifier (YAPC)

UPDATE HISTORY:
    Written 10/2021
"""
import os
import re
import shutil
import logging

def build_logger(name, **kwargs):
    """
    Builds a logging instance with the specified name

    Parameters
    ----------
    name: str
        name of the logger
    format: str
        event description message format
    level: int or obj, default logging.CRITICAL
        lowest-severity log message logger will handle
    propagate: bool, default False
        events logged will be passed to higher level handlers
    stream: obj or NoneType, default None
        specified stream to initialize StreamHandler
    """
    # set default arguments
    kwargs.setdefault('format', '%(levelname)s:%(name)s:%(message)s')
    kwargs.setdefault('level', logging.CRITICAL)
    kwargs.setdefault('propagate',False)
    kwargs.setdefault('stream',None)
    # build logger
    logger = logging.getLogger(name)
    logger.setLevel(kwargs['level'])
    logger.propagate = kwargs['propagate']
    # create and add handlers to logger
    if not logger.handlers:
        # create handler for logger
        handler = logging.StreamHandler(stream=kwargs['stream'])
        formatter = logging.Formatter(kwargs['format'])
        handler.setFormatter(formatter)
        # add handler to logger
        logger.addHandler(handler)
    return logger

# PURPOSE: convert file lines to arguments
def convert_arg_line_to_args(arg_line):
    """
    Convert file lines to arguments

    Parameters
    ----------
    arg_line: str
        line string containing a single argument and/or comments
    """
    # remove commented lines and after argument comments
    for arg in re.sub(r'\#(.*?)$',r'',arg_line).split():
        if not arg.strip():
            continue
        yield arg

# PURPOSE: make a copy of a file with all system information
def copy(source, destination, move=False):
    """
    Copy or move a file with all system information

    Parameters
    ----------
    source: str
        source file
    destination: str
        copied destination file
    move: bool, default False
        remove the source file
    """
    source = os.path.abspath(os.path.expanduser(source))
    destination = os.path.abspath(os.path.expanduser(destination))
    shutil.copyfile(source, destination)
    shutil.copystat(source, destination)
    if move:
        os.remove(source)



#!/usr/bin/env python
u"""
version.py (09/2021)
"""
from pkg_resources import get_distribution
# get version for distribution
version = get_distribution("pyYAPC").version
# append "v" before the version
full_version = "v{0}".format(version)



import os
import sys
import logging
from setuptools import setup, Extension, find_packages

logging.basicConfig(stream=sys.stderr, level=logging.INFO)
log = logging.getLogger()

# package description and keywords
description = ('Python version of the NASA GSFC YAPC '
    '("Yet Another Photon Classifier") Program')
keywords = 'ICESat-2 laser altimetry, ATLAS, photon classification, kNN filtering'
# get long_description from README.rst
with open("README.rst", mode="r", encoding="utf8") as fh:
    long_description = fh.read()
long_description_content_type = "text/x-rst"

# get install requirements
with open('requirements.txt', mode="r", encoding="utf8") as fh:
    install_requires = [line.split().pop(0) for line in fh.read().splitlines()]

# get version
with open('version.txt', mode="r", encoding="utf8") as fh:
    version = fh.read()

# list of all scripts to be included with package
scripts=[os.path.join('scripts',f) for f in os.listdir('scripts') if f.endswith('.py')]

# Setuptools 18.0 properly handles Cython extensions.
setup_requires=[
    'setuptools>=18.0',
    'cython',
]
# cythonize extensions
ext_modules=[
    Extension('yapc._dist_metrics', sources=['yapc/_dist_metrics.pyx'])
]

setup(
    name='pyYAPC',
    version=version,
    description=description,
    long_description=long_description,
    long_description_content_type=long_description_content_type,
    url='https://github.com/tsutterley/yapc',
    author='Tyler Sutterley',
    author_email='tsutterl@uw.edu',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Physics',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    keywords=keywords,
    packages=find_packages(),
    install_requires=install_requires,
    setup_requires=setup_requires,
    scripts=scripts,
    include_package_data=True,
    ext_modules=ext_modules,
)



#!python
#cython: boundscheck=False
#cython: wraparound=False
#cython: cdivision=True

# PURPOSE: create distance metric for windowed classifier
def windowed_manhattan(u, v, window=None, w=None):
    """
    Calculate distances between two collections of points
       using a windowed Manhattan metric

    Arguments
    ---------
    u: float
        First collection of coordinates
    v: float
        Second collection of coordinates
    window: float or NoneType, default None
        Distance window for reducing neighbors
    w: float or list, default []
        weights for each value
    """
    # calculate manhattan (rectilinear) distances
    cdef double d = 0.0
    d = dist(u, v, window, w)
    return d



#!/usr/bin/env python
u"""
append_YAPC_ICESat2_ATL03.py
Written by Aimee Gibbons and Tyler Sutterley (06/2022)
Reads ICESat-2 ATL03 data files and appends photon classification flags
    from YAPC (Yet Another Photon Classifier) to granules

CALLING SEQUENCE:
    python append_YAPC_ICESat2_ATL03.py ATL03_file

COMMAND LINE OPTIONS:
    --K X, -k X: number of values for KNN algorithm
        Use 0 for dynamic selection of neighbors
    --min-knn: Minimum value of K used in the KNN algorithm
    --min-ph X: minimum number of photons for a segment to be valid
    --min-x-spread X: minimum along-track spread of photon events
    --min-h-spread X: minimum window of heights for photon events
    --win_x X: along-track length of window
    --win_h X: height of window
    --aspect X: aspect ratio of x and h window
        Use 0 for pre-defined window dimensions
    --method X: algorithm for computing photon event weights
        `'ball_tree'`: use scikit.learn.BallTree with custom distance metric
        `'linear'`: use a brute-force approach with linear algebra
        `'brute'`: use a brute-force approach
    --metric X: metric for computing distances
        `'height'`: height differences
        `'manhattan'`: manhattan distances
    -O X, --output X: output file type
        append: add photon classification flags to original ATL03 file
        copy: add photon classification flags to a copied ATL03 file
        reduce: create a new file with the photon classification flags
    -V, --verbose: Verbose output to track progress
    -M X, --mode X: Permission mode of files created

PYTHON DEPENDENCIES:
    numpy: Scientific Computing Tools For Python
        https://numpy.org
        https://numpy.org/doc/stable/user/numpy-for-matlab-users.html
    scipy: Scientific Tools for Python
        https://docs.scipy.org/doc/
    h5py: Python interface for Hierarchal Data Format 5 (HDF5)
        https://h5py.org
    scikit-learn: Machine Learning in Python
        http://scikit-learn.org/stable/index.html
        https://github.com/scikit-learn/scikit-learn

PROGRAM DEPENDENCIES:
    classify_photons.py: Yet Another Photon Classifier for Geolocated Photon Data

UPDATE HISTORY:
    Updated 06/2022: can normalize weights over entire ATL03 granule
        added option for setting the minimum KNN value
    Updated 05/2022: use argparse descriptions within sphinx documentation
    Updated 04/2022: calculate weights for each ATL03 segment
    Updated 10/2021: using python logging for handling verbose output
        do not use possible TEP photons in YAPC calculation
        using new YAPC HDF5 variable names to match ASAS version
        added parsing for converting file lines to arguments
        check that given telemetry bands contain at least 1 DEM value
    Updated 09/2021: added more YAPC options for photon distance classifier
        create YAPC group and output major frame variables to HDF5
        add output options for creating copied or reduced files
    Updated 08/2021: using YAPC HDF5 variable names to match ASAS version
    Written 05/2021
"""
from __future__ import print_function

import os
import re
import h5py
import yapc
import logging
import argparse
import numpy as np

# PURPOSE: finds the vertical extent of the photons
def h_extent(h_ph):
  """Find the vertical extent of the photons
  """
  # Calculate bin edges needed for 1 meter bins
  n_bins = np.ceil(np.ptp(h_ph))
  if (n_bins <= 0):
      return 0
  #  construct a histogram
  hist,_ = np.histogram(h_ph, bins=int(n_bins))
  # Return the number of nonzero bins.
  return np.count_nonzero(hist)

# PURPOSE: reads ICESat-2 ATL03 HDF5 files
# computes photon classifications heights over 20m segments
def append_YAPC_ICESat2_ATL03(input_file, output='append', verbose=False,
    mode=0o775, **kwargs):
    # set default keyword arguments for photon classification
    kwargs.setdefault('K',0)
    kwargs.setdefault('min_knn', 5)
    kwargs.setdefault('min_ph', 3)
    kwargs.setdefault('min_xspread', 1.0)
    kwargs.setdefault('min_hspread', 0.01)
    kwargs.setdefault('win_x', 15.0)
    kwargs.setdefault('win_h', 6.0)
    kwargs.setdefault('aspect', 0.0)
    kwargs.setdefault('method', 'linear')
    kwargs.setdefault('metric', 'height')
    kwargs.setdefault('norm', 'segment')
    kwargs.setdefault('return_window', True)
    kwargs.setdefault('return_K', True)

    # create logger for verbosity level
    loglevel = logging.INFO if verbose else logging.CRITICAL
    logger = yapc.utilities.build_logger('yapc',level=loglevel)

    # compile regular expression operator for extracting data from ATL03 files
    rx = re.compile(r'(processed_)?(ATL\d+)_(\d{4})(\d{2})(\d{2})(\d{2})(\d{2})'
        r'(\d{2})_(\d{4})(\d{2})(\d{2})_(\d{3})_(\d{2})(.*?).h5$')
    # read ICESat-2 ATL03 HDF5 files (extract base parameters)
    SUB,PRD,YY,MM,DD,HH,MN,SS,TRK,CYCL,GRAN,RL,VERS,AUX=rx.findall(input_file).pop()

    # input/output directory
    directory = os.path.dirname(input_file)
    # Open the input HDF5 file for reading or appending
    assert output in ('append','copy','reduce')
    clobber = 'a' if (output == 'append') else 'r'
    f_in = h5py.File(input_file, mode=clobber)

    # output information for file
    logger.info('{0} -->'.format(input_file))

    # attributes for the output variables
    attrs = {}
    # width in meters of the selection window
    attrs['win_x'] = {}
    attrs['win_x']['units'] = "meters"
    attrs['win_x']['long_name'] = "Window width"
    attrs['win_x']['description'] = "Width in meters of the selection window"
    attrs['win_x']['source'] = "YAPC"
    attrs['win_x']['contentType'] = "referenceInformation"
    attrs['win_x']['coordinates'] = "delta_time"
    # height in meters of the selection window
    attrs['win_h'] = {}
    attrs['win_h']['units'] = "meters"
    attrs['win_h']['long_name'] = "Window height"
    attrs['win_h']['description'] = "Height in meters of the selection window"
    attrs['win_h']['source'] = "YAPC"
    attrs['win_h']['contentType'] = "referenceInformation"
    attrs['win_h']['coordinates'] = "delta_time"
    # height extent
    attrs['h_win_width'] = {}
    attrs['h_win_width']['units'] = "meters"
    attrs['h_win_width']['long_name'] = "Height extent"
    attrs['h_win_width']['description'] = "Vertical extent of the photons"
    attrs['h_win_width']['source'] = "YAPC"
    attrs['h_win_width']['contentType'] = "referenceInformation"
    attrs['h_win_width']['coordinates'] = "delta_time"
    # dynamically selected number of neighbors
    attrs['K'] = {}
    attrs['K']['units'] = "1"
    attrs['K']['long_name'] = "K"
    attrs['K']['description'] = "Dynamically selected number of neighbors"
    attrs['K']['source'] = "YAPC"
    attrs['K']['valid_min'] = kwargs['min_knn']
    attrs['K']['contentType'] = "referenceInformation"
    attrs['K']['coordinates'] = "delta_time"
    # normalization for photon event weights
    attrs['weight_ph_norm'] = {}
    attrs['weight_ph_norm']['units'] = 1
    attrs['weight_ph_norm']['long_name'] = "Maximum Weight"
    attrs['weight_ph_norm']['description'] = ("Maximum weight from the photon "
        "event classifier used as normalization for calculating the"
        "signal-to-noise ratio")
    attrs['weight_ph_norm']['source'] = "YAPC"
    attrs['weight_ph_norm']['normalization'] = kwargs['norm']
    attrs['weight_ph_norm']['contentType'] = "qualityInformation"
    attrs['weight_ph_norm']['coordinates'] = ("delta_time reference_photon_lat "
        "reference_photon_lon")
    # signal-to-noise ratio for each photon
    attrs['weight_ph'] = {}
    attrs['weight_ph']['units'] = 100
    attrs['weight_ph']['long_name'] = "Signal-to-Noise Ratio"
    attrs['weight_ph']['description'] = ("Signal-to-Noise ratio calculated using "
        "the photon event classifier, normalized using the maximum weight "
        "in an ATL03 segment")
    attrs['weight_ph']['source'] = "YAPC"
    attrs['weight_ph']['normalization'] = kwargs['norm']
    attrs['weight_ph']['contentType'] = "qualityInformation"
    attrs['weight_ph']['coordinates'] = "delta_time lat_ph lon_ph"
    # photon signal-to-noise confidence from photon classifier
    BACKG,L_CONF,M_CONF,H_CONF = (0.0,25.0,60.0,80.0)
    attrs['yapc_conf'] = {}
    attrs['yapc_conf']['units'] = 1
    attrs['yapc_conf']['valid_min'] = 0
    attrs['yapc_conf']['valid_max'] = 4
    attrs['yapc_conf']['flag_values'] = [0,2,3,4]
    attrs['yapc_conf']['confidences'] = [BACKG,L_CONF,M_CONF,H_CONF]
    attrs['yapc_conf']['flag_meanings'] = "noise low medium high"
    attrs['yapc_conf']['long_name'] = "Photon Signal Confidence"
    attrs['yapc_conf']['description'] = ("Confidence level associated with "
        "each photon event selected as signal from the photon classifier "
        "(0=noise; 2=low; 3=med; 4=high).")
    attrs['yapc_conf']['source'] = "YAPC"
    attrs['yapc_conf']['normalization'] = kwargs['norm']
    attrs['yapc_conf']['contentType'] = "qualityInformation"
    attrs['yapc_conf']['coordinates'] = "delta_time lat_ph lon_ph"

    # read each input beam within the file
    IS2_atl03_beams = []
    for gtx in [k for k in f_in.keys() if bool(re.match(r'gt\d[lr]',k))]:
        # check if subsetted beam contains data
        # check in both the geolocation and heights groups
        try:
            f_in[gtx]['geolocation']['segment_id']
            f_in[gtx]['heights']['delta_time']
        except KeyError:
            pass
        else:
            IS2_atl03_beams.append(gtx)

    # open output file
    if (output == 'append'):
        # copy filename and file object for output variables
        output_file = input_file
        f_out = f_in
    elif (output == 'copy'):
        # Copy input file to a new appended ATL03 file
        fargs=(SUB,PRD,YY,MM,DD,HH,MN,SS,TRK,CYCL,GRAN,RL,VERS,AUX)
        file_format='{0}{1}_{2}{3}{4}{5}{6}{7}_{8}{9}{10}_{11}_{12}{13}_YAPC.h5'
        output_file = os.path.join(directory,file_format.format(*fargs))
        yapc.utilities.copy(input_file,output_file)
        # Open the output HDF5 files for appending
        f_out = h5py.File(output_file, 'a')
    elif (output == 'reduce'):
        # create a new reduced file with only YAPC parameters
        fargs=(SUB,PRD,YY,MM,DD,HH,MN,SS,TRK,CYCL,GRAN,RL,VERS,AUX)
        file_format='{0}{1}_YAPC_{2}{3}{4}{5}{6}{7}_{8}{9}{10}_{11}_{12}{13}.h5'
        output_file = os.path.join(directory,file_format.format(*fargs))
        # Open the output HDF5 files for writing
        f_out = h5py.File(output_file, 'w')
        # copy file-level attributes from input to output
        for att_name,att_val in f_in.attrs.items():
            f_out.attrs[att_name] = att_val
        # for each valid beam group in the input file
        for gtx in IS2_atl03_beams:
            copy_ATL03_beam_group(f_in, f_out, gtx)

    # print output file if verbose
    logger.info('{0} ({1})'.format(output_file, output))

    # for each input beam within the file
    for gtx in sorted(IS2_atl03_beams):
        logger.info(gtx)
        # ATL03 Segment ID
        Segment_ID = f_in[gtx]['geolocation']['segment_id'][:]
        # number of ATL03 20 meter segments
        n_seg = len(Segment_ID)
        # number of photon events
        n_pe, = f_in[gtx]['heights']['delta_time'].shape

        # first photon in the segment (convert to 0-based indexing)
        Segment_Index_begin = f_in[gtx]['geolocation']['ph_index_beg'][:] - 1
        # number of photon events in the segment
        Segment_PE_count = f_in[gtx]['geolocation']['segment_ph_cnt'][:]
        # along-track distance for each ATL03 segment
        Segment_Distance = f_in[gtx]['geolocation']['segment_dist_x'][:]
        # along-track and across-track distance for photon events
        x_atc = f_in[gtx]['heights']['dist_ph_along'][:].copy()
        # photon event heights
        h_ph = f_in[gtx]['heights']['h_ph'][:].copy()
        # flag denoting photon events as possible TEP
        if (int(RL) < 4):
            isTEP = np.any((f_in[gtx]['heights']['signal_conf_ph'][:]==-2),axis=1)
        else:
            isTEP = (f_in[gtx]['heights']['quality_ph'][:] == 3)

        # photon event weights and normalization
        pe_weights = np.zeros((n_pe))
        weight_ph_norm = np.zeros((n_seg))
        # photon event variables
        heights = {}
        # photon signal-to-noise ratios from classifier
        heights['weight_ph'] = np.zeros((n_pe),dtype=np.uint8)
        # photon confidence levels from classifier
        heights['yapc_conf'] = np.zeros((n_pe),dtype=np.int8)
        # output YAPC segment variables
        yapc_window = {}
        # segment maximum range of heights
        yapc_window['h_win_width'] = np.zeros((n_seg),dtype=np.int64)
        # selection window sizes
        yapc_window['win_x'] = np.zeros((n_seg))
        yapc_window['win_h'] = np.zeros((n_seg))
        # dynamic number of values in KNN algorithm
        yapc_window['K'] = np.zeros((n_seg))
        # photon weight normalization scaled to byte
        yapc_window['weight_ph_norm'] = np.zeros((n_seg),dtype=np.uint8)
        # calculate weights for each 20m segment
        for j,_ in enumerate(Segment_ID):
            # index for 20m segment j
            idx = Segment_Index_begin[j]
            # skip segments with no photon events
            if (idx < 0):
                continue
            # number of photons in 20m segment
            cnt = Segment_PE_count[j]
            # buffer photons with previous and following segments
            cnt_m1 = 0 if (j == 0) else Segment_PE_count[j-1]
            cnt_p1 = 0 if (j == (n_seg-1)) else Segment_PE_count[j+1]
            # photon indices for segment (buffered by 1 on each side)
            i1 = np.arange(idx-cnt_m1, idx+cnt+cnt_p1, 1)
            # indices of non-TEP photons in central segment
            i2 = np.nonzero((i1 >= idx) & (i1 < (idx+cnt)) &
                np.logical_not(isTEP[i1]))
            # skip segments that are all TEP photons
            if not np.any(i2):
                continue
            # add segment distance to along-track coordinates
            distance_along_X = np.copy(x_atc[i1])
            distance_along_X += Segment_Distance[j]
            # calculate the height extent
            h_win_width = np.max([1., h_extent(h_ph[i1])])
            yapc_window['h_win_width'][j] = h_win_width.astype(np.int64)
            # photon event weights
            pe_weights[i1[i2]], win_x, win_h, K = yapc.classify_photons(
                distance_along_X, h_ph[i1], h_win_width, i2, **kwargs)
            # selection window sizes (can be dynamically calculated)
            yapc_window['win_x'][j] = np.copy(win_x)
            yapc_window['win_h'][j] = np.copy(win_h)
            # dynamically selected number of neighbors
            yapc_window['K'][j] = np.copy(K)
            # normalize photon weights for each segment
            if (kwargs['norm'] == 'segment'):
                weight_ph_norm[j] = np.max(pe_weights[i1[i2]])

        # normalize photon weights over complete granule
        if (kwargs['norm'] == 'granule'):
            weight_ph_norm[:] = np.max(pe_weights)

        # calculated scaled weights for each 20m segment
        for j,_ in enumerate(Segment_ID):
            # index for 20m segment j
            idx = Segment_Index_begin[j]
            # skip segments with no photon events
            if (idx < 0):
                continue
            # number of photons in 20m segment
            cnt = Segment_PE_count[j]
            # buffer photons with previous and following segments
            cnt_m1 = 0 if (j == 0) else Segment_PE_count[j-1]
            cnt_p1 = 0 if (j == (n_seg-1)) else Segment_PE_count[j+1]
            # photon indices for segment (buffered by 1 on each side)
            i1 = np.arange(idx-cnt_m1, idx+cnt+cnt_p1, 1)
            # indices of non-TEP photons in central segment
            i2 = np.nonzero((i1 >= idx) & (i1 < (idx+cnt)) &
                np.logical_not(isTEP[i1]))
            # skip segments that are all TEP photons
            if not np.any(i2):
                continue
            # skip segments where the maximum weight is zero
            if (weight_ph_norm[j] == 0):
                continue
            # scaled photon event weights from photon classifier
            segment_weights = 255.0*(pe_weights[i1[i2]]/weight_ph_norm[j])
            # verify segment weights and copy to output
            np.clip(segment_weights, 0, 255, out=segment_weights)
            heights['weight_ph'][i1[i2]] = segment_weights.astype(np.uint8)
            # calculate normalization for 20m segment
            yapc_window['weight_ph_norm'][j] = weight_ph_norm[j].astype(np.uint8)
            # photon event signal-to-noise ratio from photon classifier
            if (weight_ph_norm[j] > 0):
                # calculate PE signal-to-noise ratio
                scaled_SNR = (100.0*segment_weights)/weight_ph_norm[j]
                # verify PE SNR values and add to output array
                np.clip(scaled_SNR, 0, 100, out=scaled_SNR)
                # calculate confidence levels from photon classifier
                segment_class = np.zeros_like(scaled_SNR, dtype=np.int8)
                segment_class[scaled_SNR >= L_CONF] = 2
                segment_class[scaled_SNR >= M_CONF] = 3
                segment_class[scaled_SNR >= H_CONF] = 4
                # copy segment classification to output heights variable
                heights['yapc_conf'][i1[i2]] = segment_class.copy()

        # add YAPC attributes to geolocation group
        f_out[gtx]['geolocation'].attrs['yapc_version'] = \
            yapc.version.full_version
        for att_name in ['min_ph','min_xspread','min_hspread','metric']:
            f_out[gtx]['geolocation'].attrs[att_name] = kwargs[att_name]

        # segment variables from photon classifier
        for key in ['h_win_width','win_x','win_h','K','weight_ph_norm']:
            val = '{0}/{1}/{2}'.format(gtx,'geolocation',key)
            logger.info(val)
            h5 = f_out.create_dataset(val, np.shape(yapc_window[key]),
                data=yapc_window[key], dtype=yapc_window[key].dtype,
                compression='gzip')
            # attach dimensions
            for i,dim in enumerate(['delta_time']):
                h5.dims[i].attach_scale(f_out[gtx]['geolocation'][dim])
            # add HDF5 variable attributes
            for att_name,att_val in attrs[key].items():
                h5.attrs[att_name] = att_val

        # photon signal-to-noise ratio from photon classifier
        for key in ['weight_ph','yapc_conf']:
            val = '{0}/{1}/{2}'.format(gtx,'heights',key)
            logger.info(val)
            h5 = f_out.create_dataset(val, np.shape(heights[key]),
                data=heights[key], dtype=heights[key].dtype,
                compression='gzip')
            # attach dimensions
            for i,dim in enumerate(['delta_time']):
                h5.dims[i].attach_scale(f_out[gtx]['heights'][dim])
            # add HDF5 variable attributes
            for att_name,att_val in attrs[key].items():
                h5.attrs[att_name] = att_val

    # close the HDF5 files
    f_in.close()
    f_out.close()
    # change the permissions mode
    os.chmod(output_file, mode=mode)

# PURPOSE: copy ATL03 variables and attributes to output file
def copy_ATL03_beam_group(f_in, f_out, gtx):
    # group variables to copy for each beam group
    groups = {}
    groups['geolocation'] = ['delta_time','segment_id',
        'reference_photon_lat','reference_photon_lon']
    groups['heights'] = ['delta_time','lat_ph','lon_ph']
    # attributes to not copy from input HDF5 file
    invalid_attributes = ['CLASS','DIMENSION_LIST','NAME','REFERENCE_LIST']
    # create the beam group
    f_out.create_group(gtx)
    # copy group attributes from input to output
    for att_name,att_val in f_in[gtx].attrs.items():
        f_out[gtx].attrs[att_name] = att_val
    # create data groups and copy necessary variables
    for group,keys in groups.items():
        f_out[gtx].create_group(group)
        # for each variable to copy from the input file
        for key in keys:
            val = '{0}/{1}/{2}'.format(gtx,group,key)
            h5 = f_out.create_dataset(val, data=f_in[val],
                dtype=f_in[val].dtype, compression='gzip')
            # create or attach dimensions
            if (key == 'delta_time'):
                # make dimension
                h5.make_scale('delta_time')
            else:
                # attach dimensions
                for i,dim in enumerate(['delta_time']):
                    h5.dims[i].attach_scale(f_out[gtx][group][dim])
            # copy group attributes from input to output
            for att_name,att_val in f_in[val].attrs.items():
                if att_name not in invalid_attributes:
                    f_out[val].attrs[att_name] = att_val

# PURPOSE: create argument parser
def arguments():
    # Read the system arguments listed after the program
    parser = argparse.ArgumentParser(
        description="""Reads ICESat-2 ATL03 data files and appends
            photon classification flags from YAPC
            """,
        fromfile_prefix_chars="@"
    )
    parser.convert_arg_line_to_args = yapc.utilities.convert_arg_line_to_args
    # command line parameters
    parser.add_argument('infile',
        type=lambda p: os.path.abspath(os.path.expanduser(p)), nargs='+',
        help='ICESat-2 ATL03 file to run')
    # number of values for KNN algorithm
    parser.add_argument('--K','-k',
        type=int, default=0,
        help='Number of values for KNN algorithm')
    # minimum value of K for KNN algorithm
    parser.add_argument('--min-knn',
        type=int, default=5,
        help='Minimum value of K used in the KNN algorithm')
    # minimum number of photons for a segment
    parser.add_argument('--min-ph',
        type=int, default=3,
        help='Minimum number of photons for a segment to be valid')
    # minimum along-track spread of photon events
    parser.add_argument('--min-x-spread',
        type=float, default=1.0,
        help='Minimum along-track spread of photon events')
    # minimum window of heights for photon events
    parser.add_argument('--min-h-spread',
        type=float, default=0.01,
        help='Minimum window of heights for photon events')
    # x and h window
    parser.add_argument('--win-x',
        type=float, default=15.0,
        help='Along-track length of window')
    parser.add_argument('--win-h',
        type=float, default=6.0,
        help='Height of window')
    # aspect ratio of x and h window
    parser.add_argument('--aspect',
        type=float, default=0.0,
        help='Aspect ratio of x and h window')
    # algorithm for computing photon event weights
    choices = ('ball_tree','linear','brute')
    parser.add_argument('--method',
        type=str.lower, default='linear', choices=choices,
        help='Algorithm for computing photon event weights')
    # metric for computing distances
    choices = ('height','manhattan')
    parser.add_argument('--metric',
        type=str.lower, default='height', choices=choices,
        help='Metric for computing distances')
    # normalization scheme for weights (segment level or granule)
    choices = ('segment','granule')
    parser.add_argument('--norm',
        type=str.lower, default='segment', choices=choices,
        help='Normalization scheme for weights')
    # output file type (appended, copied or reduced)
    choices = ('append','copy','reduce')
    parser.add_argument('--output','-O',
        type=str.lower, default='copy', choices=choices,
        help='Output file type')
    # verbosity settings
    parser.add_argument('--verbose','-V',
        default=False, action='store_true',
        help='Verbose output of run')
    # permissions mode of the local files (number in octal)
    parser.add_argument('--mode','-M',
        type=lambda x: int(x,base=8), default=0o775,
        help='Permissions mode of output files')
    # return the parser
    return parser

# This is the main part of the program that calls the individual functions
def main():
    # Read the system arguments listed after the program
    parser = arguments()
    args = parser.parse_args()

    # run the program for the ATL03 file
    for ATL03_file in args.infile:
        append_YAPC_ICESat2_ATL03(ATL03_file,
            K=args.K,
            min_knn=args.min_knn,
            min_ph=args.min_ph,
            min_xspread=args.min_x_spread,
            min_hspread=args.min_h_spread,
            win_x=args.win_x,
            win_h=args.win_h,
            aspect=args.aspect,
            method=args.method,
            metric=args.metric,
            norm=args.norm,
            output=args.output,
            verbose=args.verbose,
            mode=args.mode)

# run main program
if __name__ == '__main__':
    main()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
# import sys
import datetime
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = 'pyYAPC'
year = datetime.date.today().year
copyright = f"2021\u2013{year}, Tyler C. Sutterley"
author = 'Aimee Gibbons and Tyler Sutterley'

# The full version, including alpha/beta/rc tags
with open(os.path.abspath('../../version.txt')) as fh:
    release = fh.read()

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "numpydoc",
    "sphinx.ext.graphviz",
    "sphinx.ext.viewcode",
    "sphinxarg.ext"
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['**.ipynb_checkpoints']

# location of master document (by default sphinx looks for contents.rst)
master_doc = 'index'

# -- Configuration options ---------------------------------------------------
autosummary_generate = True
numfig = True
autodoc_member_order = 'bysource'
numpydoc_show_class_members = False
pygments_style = 'native'

# -- Options for HTML output -------------------------------------------------

# html_title = "yapc"
html_short_title = "yapc"
html_show_sourcelink = False
html_show_sphinx = True
html_show_copyright = True

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'
html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']
repository_url = f"https://github.com/tsutterley/yapc"
html_context = {
    "menu_links": [
        (
            '<i class="fa fa-github fa-fw"></i> Source Code',
            repository_url,
        ),
        (
            '<i class="fa fa-book fa-fw"></i> License',
            f"{repository_url}/blob/main/LICENSE",
        ),
    ],
}

# Load the custom CSS files (needs sphinx >= 1.6 for this to work)
def setup(app):
    app.add_css_file("style.css")
